package broker

import (
	"context"
	"encoding/json"
	"fmt"
	"log"

	"github.com/pborman/uuid"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/boshclient"
	"github.com/pivotal-cf/on-demand-service-broker/brokercontext"
)

func (b *Broker) Deprovision(
	ctx context.Context,
	instanceID string,
	_ brokerapi.DeprovisionDetails,
	asyncAllowed bool,
) (brokerapi.DeprovisionServiceSpec, error) {
	b.deploymentLock.Lock()
	defer b.deploymentLock.Unlock()

	requestID := uuid.New()
	ctx = brokercontext.New(ctx, string(OperationTypeDelete), requestID, b.serviceOffering.Name, instanceID)
	logger := b.loggerFactory.NewLoggerWithContext(ctx)

	if !asyncAllowed {
		return brokerapi.DeprovisionServiceSpec{}, brokerapi.ErrAsyncRequired
	}

	instanceState, err := b.cfClient.GetInstanceState(instanceID, logger)
	if err != nil {
		return deprovisionErr(NewGenericError(ctx, err), logger)
	}

	plan, found := b.serviceOffering.FindPlanByID(instanceState.PlanID)
	if found {
		if errand := plan.PreDeleteErrand(); errand != "" {
			return b.runPreDeleteErrand(ctx, instanceID, errand, logger)
		}
	}

	return b.deleteInstance(ctx, instanceID, plan, logger)
}

func (b *Broker) runPreDeleteErrand(
	ctx context.Context,
	instanceID string,
	preDeleteErrand string,
	logger *log.Logger,
) (brokerapi.DeprovisionServiceSpec, error) {
	logger.Printf("running pre-delete errand for instance %s\n", instanceID)

	boshContextID := uuid.New()

	taskID, err := b.boshClient.RunErrand(
		deploymentName(instanceID),
		preDeleteErrand,
		boshContextID,
		logger,
	)
	if err != nil {
		return deprovisionErr(NewGenericError(ctx, err), logger)
	}

	operationData, err := json.Marshal(OperationData{
		OperationType: OperationTypeDelete,
		BoshTaskID:    taskID,
		BoshContextID: boshContextID,
	})

	if err != nil {
		return deprovisionErr(NewGenericError(ctx, err), logger)
	}

	return brokerapi.DeprovisionServiceSpec{IsAsync: true, OperationData: string(operationData)}, nil
}

func (b *Broker) deleteInstance(
	ctx context.Context,
	instanceID string,
	planConfig Plan,
	logger *log.Logger,
) (brokerapi.DeprovisionServiceSpec, error) {
	logger.Printf("deleting deployment for instance %s\n", instanceID)
	taskID, err := b.boshClient.DeleteDeployment(deploymentName(instanceID), "", logger)
	switch err.(type) {
	case boshclient.DeploymentNotFoundError:
		return deprovisionErr(NewDisplayableError(
			brokerapi.ErrInstanceDoesNotExist,
			fmt.Errorf("error deprovisioning: instance %s, not found: %s", instanceID, err),
		), logger)
	case boshclient.RequestError:
		return deprovisionErr(NewBoshRequestError("delete", err), logger)
	case error:
		return deprovisionErr(NewGenericError(
			ctx,
			fmt.Errorf("error deprovisioning: deleting bosh deployment: %s", err),
		), logger)
	}

	logger.Printf("Bosh task id for Delete instance %s was %d\n", instanceID, taskID)
	ctx = brokercontext.WithBoshTaskID(ctx, taskID)

	operationData, err := json.Marshal(OperationData{
		OperationType: OperationTypeDelete,
		BoshTaskID:    taskID,
	})

	if err != nil {
		return deprovisionErr(NewGenericError(ctx, err), logger)
	}

	return brokerapi.DeprovisionServiceSpec{
		IsAsync:       true,
		OperationData: string(operationData),
	}, nil
}

func deprovisionErr(
	err DisplayableError,
	logger *log.Logger,
) (brokerapi.DeprovisionServiceSpec, error) {
	logger.Println(err)
	return brokerapi.DeprovisionServiceSpec{IsAsync: true}, err.ErrorForCFUser()
}
